import hashlib
import json
from typing import Any, Dict, List


def _create_json_hash(data: Any, hash_function: str) -> str:
    """Create an sha256 hash from json-converted data, sorted by key names."""
    stringified = json.dumps(
        data,
        allow_nan=False,
        ensure_ascii=False,
        indent=None,
        separators=(",", ":"),
        skipkeys=False,
        sort_keys=True,
    )
    if hash_function == "sha256":
        m = hashlib.sha256()
    if hash_function == "sha384":
        m = hashlib.sha384()
    if hash_function == "sha512":
        m = hashlib.sha512()
    m.update(stringified.encode("utf-8"))

    return m.hexdigest()


def _create_sorted_hash_list(data: List, hash_function: str) -> List[str]:
    """Create a sorted sha256 hash list."""
    out = []
    for obj in data:
        hash = _create_json_hash(obj, hash_function=hash_function)
        out.append(hash)
    out.sort()
    return out


def _build_path(key: str, base_path: str):
    """Build a path string."""
    if base_path:
        return f"{base_path}|{key}"
    return key


def _build_element(path: str, siblings: List[str], value: Any):
    """Build an element dictionary based on presence of sibling data."""
    if siblings:
        return {
            "path": path,
            "siblings": siblings,
            "value": value,
        }

    return {
        "path": path,
        "value": value,
    }


def _flatten_json(data: Any, hash_function: str, path: str = "", siblings: List = [], debug: bool = False) -> List:
    """Flatten json data structures into a sibling-aware data element list."""
    out = []

    # Process non-empty dicts
    if type(data) is dict and data:
        for key in data.keys():
            p = _build_path(key=f"{{{key}}}", base_path=path)
            output = _flatten_json(
                data=data[key],
                hash_function=hash_function,
                path=p,
                siblings=siblings,
                debug=debug,
            )
            out.extend(output)
        return out

    # Process non-empty lists
    if type(data) is list and data:
        p = _build_path(key=f"[{len(data)}]", base_path=path)

        # Iterate and collect sibling structures, which will be then attached to each sibling element
        siblings = []
        for item in data:
            output = _flatten_json(data=item, hash_function=hash_function, path=p, debug=debug)
            siblings.extend(output)

        # Debug mode, which allows non-hashed sibling structures to be inspected and tested against
        if not debug:
            siblings = _create_sorted_hash_list(data=siblings, hash_function=hash_function)
            siblings = _create_json_hash(data=siblings, hash_function=hash_function)

        # Recurse with each value in list to typecheck it and eventually get the element value
        for item in data:
            output = _flatten_json(
                data=item,
                hash_function=hash_function,
                path=p,
                siblings=siblings,
                debug=debug,
            )
            out.extend(output)
        return out

    element = _build_element(path=path, siblings=siblings, value=data)
    out.append(element)
    return out


def _create_jfpv1_fingerprint(data: Any, hash_function: str, version: int):
    """Create a jfpv1 fingerprint."""
    flattened_json = _flatten_json(data=data, hash_function=hash_function)
    sorted_hash_list = _create_sorted_hash_list(data=flattened_json, hash_function=hash_function)
    hex_digest = _create_json_hash(data=sorted_hash_list, hash_function=hash_function)
    return f"jfpv1${hash_function}${hex_digest}"



import json

from ._exceptions import FingerprintJSONLoadError


def _load_json(data: str):
    try:
        return json.loads(data)
    except Exception:
        err = "Unable to load JSON"
        raise FingerprintJSONLoadError(err) from None



import json
import unittest

from json_fingerprint import _exceptions, create, decode


class TestDecode(unittest.TestCase):
    def test_jfpv1_decode(self):
        """Test json fingerprint decoder.

        Verify that:
        - Fingerprints of all jfpv1 SHA-2 variants are properly decoded
        - Exception is properly raised with invalid fingerprint input
        """
        input = json.dumps({"foo": "bar"})
        jfpv1_sha256 = create(input=input, hash_function="sha256", version=1)
        jfpv1_sha384 = create(input=input, hash_function="sha384", version=1)
        jfpv1_sha512 = create(input=input, hash_function="sha512", version=1)

        version, hash_function, hash = decode(fingerprint=jfpv1_sha256)
        self.assertEqual(version, 1)
        self.assertEqual(hash_function, "sha256")
        self.assertEqual(hash, jfpv1_sha256.split("$")[-1])

        version, hash_function, hash = decode(fingerprint=jfpv1_sha384)
        self.assertEqual(version, 1)
        self.assertEqual(hash_function, "sha384")
        self.assertEqual(hash, jfpv1_sha384.split("$")[-1])

        version, hash_function, hash = decode(fingerprint=jfpv1_sha512)
        self.assertEqual(version, 1)
        self.assertEqual(hash_function, "sha512")
        self.assertEqual(hash, jfpv1_sha512.split("$")[-1])

        with self.assertRaises(_exceptions.FingerprintStringFormatError):
            decode(fingerprint="invalid fingerprint")


if __name__ == "__main__":
    unittest.main()



from ._jfpv1 import _create_jfpv1_fingerprint
from ._load_json import _load_json
from ._validators import (
    _validate_hash_function,
    _validate_input_type,
    _validate_version,
)


def create(input: str, hash_function: str, version: int) -> str:
    """Create json fingerprints with the selected hash function and jfp version."""
    _validate_version(version=version)
    _validate_input_type(input=input)
    _validate_hash_function(hash_function=hash_function, version=version)
    loaded = _load_json(data=input)
    return _create_jfpv1_fingerprint(data=loaded, hash_function=hash_function, version=version)



import re

from ._exceptions import (
    FingerprintHashFunctionError,
    FingerprintInputDataTypeError,
    FingerprintStringFormatError,
    FingerprintVersionError,
)

SHA256_JFP_REGEX_PATTERN = re.compile("^jfpv1\\$sha256\\$[0-9a-f]{64}$")
SHA384_JFP_REGEX_PATTERN = re.compile("^jfpv1\\$sha384\\$[0-9a-f]{96}$")
SHA512_JFP_REGEX_PATTERN = re.compile("^jfpv1\\$sha512\\$[0-9a-f]{128}$")

JFPV1_HASH_FUNCTIONS = (
    "sha256",
    "sha384",
    "sha512",
)

JSON_FINGERPRINT_VERSIONS = (1,)


def _validate_hash_function(hash_function: str, version: int):
    if version == 1 and hash_function not in JFPV1_HASH_FUNCTIONS:
        err = f"Expected one of supported hash functions '{JFPV1_HASH_FUNCTIONS}', instead got '{hash_function}'"
        raise FingerprintHashFunctionError(err)


def _validate_input_type(input: str):
    if type(input) is not str:
        err = f'Expected data type \'{type("")}\' (JSON in string format), instead got \'{type(input)}\''
        raise FingerprintInputDataTypeError(err)


def _validate_version(version: int):
    if version not in JSON_FINGERPRINT_VERSIONS:
        err = f"Expected one of supported JSON fingerprint versions '{JSON_FINGERPRINT_VERSIONS}', instead got '{version}'"
        raise FingerprintVersionError(err)


def _validate_fingerprint_format(fingerprint: str):
    is_valid = False

    if SHA256_JFP_REGEX_PATTERN.match(fingerprint) or SHA384_JFP_REGEX_PATTERN.match(fingerprint) or SHA512_JFP_REGEX_PATTERN.match(fingerprint):
        is_valid = True

    if not is_valid:
        err = "Expected JSON fingerprint in format '{fingerprint_version}${hash_function}${hex_digest}', " f"instead got: {fingerprint}"
        raise FingerprintStringFormatError(err)



import json
import unittest

from json_fingerprint import _exceptions, _validators, create


class TestValidators(unittest.TestCase):
    def test_json_fingerprint_version_error(self):
        """Test json fingerprint version selector error

        Verify that:
        - FingerprintVersionError is properly raised with an unsupported version
        - FingerprintVersionError is not raised with a supported version
        """
        with self.assertRaises(_exceptions.FingerprintVersionError):
            create(input=json.dumps({"foo": "bar"}), hash_function="sha256", version=-1)

        try:
            _validators._validate_version(version=1)
        except Exception:
            err = "_validate_version() failed with a valid version"
            self.fail(err)

        with self.assertRaises(_exceptions.FingerprintVersionError):
            _validators._validate_version(version=-1)

    def test_input_data_type_error(self):
        """Test jfpv1 input data type error

        Verify that:
        - FingerprintInputDataTypeError is properly raised with a non-string input
        - FingerprintInputDataTypeError is not raised with a string input
        """
        with self.assertRaises(_exceptions.FingerprintInputDataTypeError):
            create(input=123, hash_function="sha256", version=1)

        try:
            _validators._validate_input_type(input="abc")
        except Exception:
            err = "_validate_input_type() failed with a valid string"
            self.fail(err)

        with self.assertRaises(_exceptions.FingerprintInputDataTypeError):
            _validators._validate_input_type(input=123)

    def test_jfpv1_hash_function_error(self):
        """Test json fingerprint hash function selector error

        Verify that:
        - FingerprintHashFunctionError is properly raised with an unsupported hash function selector
        - FingerprintHashFunctionError is not raised with a supported hash function selector
        """
        with self.assertRaises(_exceptions.FingerprintHashFunctionError):
            create(input=json.dumps({"foo": "bar"}), hash_function="not123", version=1)

        try:
            _validators._validate_hash_function(hash_function="sha256", version=1)
        except Exception:
            err = "_validate_hash_function() failed with a valid hash function and version"
            self.fail(err)

        with self.assertRaises(_exceptions.FingerprintHashFunctionError):
            _validators._validate_hash_function(hash_function="not123", version=1)

    def test_validate_fingerprint_format(self):
        """Test json fingerprint format validator.

        Verify that:
        - FingerprintStringFormatError is properly raised with invalid fingerprint format
        - FingerprintStringFormatError is not raised with a valid fingerprint
        """
        with self.assertRaises(_exceptions.FingerprintStringFormatError):
            _validators._validate_fingerprint_format(fingerprint="invalid fingerprint")

        input = json.dumps({"foo": "bar"})
        jfpv1_sha256 = create(input=input, hash_function="sha256", version=1)
        jfpv1_sha384 = create(input=input, hash_function="sha384", version=1)
        jfpv1_sha512 = create(input=input, hash_function="sha512", version=1)

        try:
            _validators._validate_fingerprint_format(fingerprint=jfpv1_sha256)
        except Exception:
            err = "_validate_fingerprint_format() failed with a valid json-sha256 fingerprint"
            self.fail(err)

        try:
            _validators._validate_fingerprint_format(fingerprint=jfpv1_sha384)
        except Exception:
            err = "_validate_fingerprint_format() failed with a valid json-sha384 fingerprint"
            self.fail(err)

        try:
            _validators._validate_fingerprint_format(fingerprint=jfpv1_sha512)
        except Exception:
            err = "_validate_fingerprint_format() failed with a valid json-sha512 fingerprint"
            self.fail(err)


if __name__ == "__main__":
    unittest.main()



import unittest

from json_fingerprint.tests.test_create import TestCreate
from json_fingerprint.tests.test_decode import TestDecode
from json_fingerprint.tests.test_find_matches import TestFindMatches
from json_fingerprint.tests.test_jfpv1 import TestJfpv1
from json_fingerprint.tests.test_match import TestMatch
from json_fingerprint.tests.test_validators import TestValidators

if __name__ == "__main__":
    unittest.main()



import json
import os
import unittest

from json_fingerprint import create
from json_fingerprint._exceptions import FingerprintJSONLoadError

TESTS_DIR = os.path.dirname(__file__)
TESTDATA_DIR = os.path.join(TESTS_DIR, "testdata")


class TestCreate(unittest.TestCase):
    def test_jfpv1_json_load_error(self):
        """Test json fingerprint raw json string load error

        Verify that:
        - FingerprintJSONLoadError is properly raised with malformed json input string
        """
        with self.assertRaises(FingerprintJSONLoadError):
            create('{"foo": bar}', hash_function="sha256", version=1)

    def test_jfpv1_sha256_output_format(self):
        """Test jfpv1 output format

        Verify that:
        - Complete jfpv1-sha256 output fingerprint is properly formatted
        """
        fp = create(input='{"foo": "bar"}', hash_function="sha256", version=1)
        self.assertRegex(fp, "^jfpv1\\$sha256\\$[0-9a-f]{64}$")

    def test_jfpv1_sha384_output_format(self):
        """Test jfpv1 output format

        Verify that:
        - Complete jfpv1-sha256 output fingerprint is properly formatted
        """
        fp = create(input='{"foo": "bar"}', hash_function="sha384", version=1)
        self.assertRegex(fp, "^jfpv1\\$sha384\\$[0-9a-f]{96}$")

    def test_jfpv1_sha512_output_format(self):
        """Test jfpv1 output format

        Verify that:
        - Complete jfpv1-sha256 output fingerprint is properly formatted
        """
        fp = create(input='{"foo": "bar"}', hash_function="sha512", version=1)
        self.assertRegex(fp, "^jfpv1\\$sha512\\$[0-9a-f]{128}$")

    def test_jfpv1_sha256_mixed_order(self):
        """Test jfpv1 sha256 mixed order fingerprint match

        Verify that:
        - The fingerprints of test objects 1 and 2 match despite same data being ordered differently
        - The fingerprints also match against a known valid fingerprint
        """
        with open(os.path.join(TESTDATA_DIR, "jfpv1_test_obj_1.json"), "r") as file:
            self.test_obj_1 = file.read()
            file.close()

        with open(os.path.join(TESTDATA_DIR, "jfpv1_test_obj_2.json"), "r") as file:
            self.test_obj_2 = file.read()
            file.close()
        fp_1 = create(self.test_obj_1, hash_function="sha256", version=1)
        fp_2 = create(self.test_obj_2, hash_function="sha256", version=1)
        self.assertEqual(fp_1, fp_2)
        self.assertEqual(fp_1, "jfpv1$sha256$b182c755347a6884fd11f1194cbe0961f548e5ac62be78a56c48c3c05eb56650")

    def test_jfpv1_sha256_structural_distinction_1(self):
        """Test jfpv1 json flattener's structural value distinction

        Verify that:
        - Identical values at identical depths, but held in different data structures,
          don't produce identical outputs
        """
        obj_in_1 = [
            1,
            [1, [2, 2]],
            [2, [2, 2]],
        ]
        fp_1 = create(input=json.dumps(obj_in_1), hash_function="sha256", version=1)

        obj_in_2 = [
            1,
            [1, 2, [2, 2, 2, 2]],
        ]
        fp_2 = create(input=json.dumps(obj_in_2), hash_function="sha256", version=1)

        self.assertNotEqual(fp_1, fp_2)

    def test_jfpv1_sha256_structural_distinction_2(self):
        """Test jfpv1 json flattener's structural value distinction

        Verify that:
        - Values in identical data structure paths, but different sibling values, don't get matched
        """
        obj_in_1 = [
            [1, ["x", "x"]],
            [2, ["y", "y"]],
        ]
        fp_1 = create(input=json.dumps(obj_in_1), hash_function="sha256", version=1)

        obj_in_2 = [
            [1, ["x", "y"]],
            [2, ["x", "y"]],
        ]
        fp_2 = create(input=json.dumps(obj_in_2), hash_function="sha256", version=1)

        self.assertNotEqual(fp_1, fp_2)

    def test_jfpv1_empty_list_as_value(self):
        """Test jfpv1 json flattener's ability to handle empty lists as values

        Versions up to 0.12.2 did not acknowledge empty lists as values.
        Related issue: https://github.com/cobaltine/json-fingerprint/issues/33

        Verify that:
        - Empty lists (and, as such, underlying data structure paths) are not ignored by the json flattener
        """
        obj_in_1 = {"field1": "yes"}
        fp_1 = create(input=json.dumps(obj_in_1), hash_function="sha256", version=1)

        obj_in_2 = {"field1": "yes", "field2": []}
        fp_2 = create(input=json.dumps(obj_in_2), hash_function="sha256", version=1)

        self.assertNotEqual(fp_1, fp_2)

    def test_jfpv1_empty_dict_as_value(self):
        """Test jfpv1 json flattener's ability to handle empty dicts as values

        Versions up to 0.12.2 did not acknowledge empty dicts as values.
        Related issue: https://github.com/cobaltine/json-fingerprint/issues/33

        Verify that:
        - Empty dicts (and, as such, underlying data structure paths) are not ignored by the json flattener"""

        obj_in_1 = {"field1": "yes"}
        fp_1 = create(input=json.dumps(obj_in_1), hash_function="sha256", version=1)

        obj_in_2 = {"field1": "yes", "field2": {}}
        fp_2 = create(input=json.dumps(obj_in_2), hash_function="sha256", version=1)

        self.assertNotEqual(fp_1, fp_2)


if __name__ == "__main__":
    unittest.main()



import hashlib
import json
import unittest

from json_fingerprint import _jfpv1


class TestJfpv1(unittest.TestCase):
    def test_create_json_hash(self):
        """Test jfpv1 json sha256 hash creation

        Verify that:
        - Key-value pairs are sorted properly and the resulting SHA256 matches the expected hash
        """
        data = {"value": "bar", "path": "{foo}"}
        stringified = json.dumps(
            data,
            allow_nan=False,
            ensure_ascii=False,
            indent=None,
            separators=(",", ":"),
            skipkeys=False,
            sort_keys=True,
        )
        m = hashlib.sha256()
        m.update(stringified.encode("utf-8"))
        expected_hex_digest = m.hexdigest()
        hex_digest = _jfpv1._create_json_hash(data=data, hash_function="sha256")
        self.assertEqual(hex_digest, expected_hex_digest)

    def test_jfpv1_create_sorted_hash_list(self):
        """Test jfpv1 hash list used for condensing unique identifiers into an easily sortable list

        Verify that:
        - The hash list produces valid SHA256 hashes from json-formatted data
        - Sorts the hashes properly
        """
        input_data = [
            # SHA256 (json-formatted): ac8d8342bbb2362d13f0a559a3621bb407011368895164b628a54f7fc33fc43c
            "a",
            # SHA256 (json-formatted): c100f95c1913f9c72fc1f4ef0847e1e723ffe0bde0b36e5f36c13f81fe8c26ed
            "b",
            # SHA256 (json-formatted): 879923da020d1533f4d8e921ea7bac61e8ba41d3c89d17a4d14e3a89c6780d5d
            "c",
        ]

        output_data_hashes = _jfpv1._create_sorted_hash_list(data=input_data, hash_function="sha256")
        input_data_hashes = []
        for datum in input_data:
            m = hashlib.sha256()
            json_formatted_input = json.dumps(datum)
            m.update(json_formatted_input.encode("utf-8"))
            input_data_hashes.append(m.hexdigest())
        input_data_hashes.sort()

        self.assertEqual(input_data_hashes, output_data_hashes)

    def test_jfpv1_build_path(self):
        """Test jfpv1 raw path formatting

        Verify that:
        - Path list and dict element encapsulation work as intended
        - Combination paths are separated properly (pipe character)
        - List and dict indicators, and path separator, display correctly even if same
          characters ([]{}|) are used in dict field names
        """
        root = ""
        dict_key = "{foo}"  # Dict with a field named 'foo'
        list_key = "[5]"  # List with 5 elements in it

        dict_root_path = _jfpv1._build_path(key=dict_key, base_path=root)
        self.assertEqual(dict_root_path, "{foo}")

        list_root_path = _jfpv1._build_path(key=list_key, base_path=root)
        self.assertEqual(list_root_path, "[5]")

        combination_path = _jfpv1._build_path(key=dict_key, base_path=list_key)
        self.assertEqual(combination_path, "[5]|{foo}")

        obj_out_raw = _jfpv1._flatten_json(data=[1, {"[1]|{foo}": "bar"}, 2], hash_function="sha256")
        self.assertEqual(obj_out_raw[0]["path"], "[3]")
        self.assertEqual(obj_out_raw[1]["path"], "[3]|{[1]|{foo}}")
        self.assertEqual(obj_out_raw[2]["path"], "[3]")

    def test_build_element(self):
        without_siblings_val = _jfpv1._build_element(path="{foo}", siblings=[], value="bar")
        self.assertEqual(without_siblings_val, {"path": "{foo}", "value": "bar"})
        with_siblings_val = _jfpv1._build_element(path="[2]", siblings="abc", value="foo")
        self.assertEqual(with_siblings_val, {"path": "[2]", "siblings": "abc", "value": "foo"})

    def test_jfpv1_flattened_json_sha256_fingerprint(self):
        """Test jfpv1 json flattener and fingerprint creation

        Verify that:
        - jfpv1 json flattener produces expected raw output format (non-hashed siblings in debug mode)
        - jfpv1 json flattener produces expected standard output format (hashed siblings)
        - jfpv1 json flattener output will produce valid jfpv1 fingerprints
        """
        obj_in = [
            1,
            [2, 3],
        ]
        obj_out_raw = _jfpv1._flatten_json(data=obj_in, hash_function="sha256", debug=True)
        expected_obj_out_raw = [
            {
                "path": "[2]",
                "siblings": [
                    {"path": "[2]", "value": 1},
                    {"path": "[2]|[2]", "siblings": [{"path": "[2]|[2]", "value": 2}, {"path": "[2]|[2]", "value": 3}], "value": 2},
                    {"path": "[2]|[2]", "siblings": [{"path": "[2]|[2]", "value": 2}, {"path": "[2]|[2]", "value": 3}], "value": 3},
                ],
                "value": 1,
            },
            {"path": "[2]|[2]", "siblings": [{"path": "[2]|[2]", "value": 2}, {"path": "[2]|[2]", "value": 3}], "value": 2},
            {"path": "[2]|[2]", "siblings": [{"path": "[2]|[2]", "value": 2}, {"path": "[2]|[2]", "value": 3}], "value": 3},
        ]
        self.assertEqual(obj_out_raw, expected_obj_out_raw)

        obj_out = _jfpv1._flatten_json(data=obj_in, hash_function="sha256")
        inner_list_siblings_hash_list = _jfpv1._create_sorted_hash_list(
            [{"path": "[2]|[2]", "value": 2}, {"path": "[2]|[2]", "value": 3}], hash_function="sha256"
        )
        siblings_1 = _jfpv1._create_sorted_hash_list(
            [
                {"path": "[2]", "value": 1},
                {"path": "[2]|[2]", "siblings": _jfpv1._create_json_hash(data=inner_list_siblings_hash_list, hash_function="sha256"), "value": 2},
                {"path": "[2]|[2]", "siblings": _jfpv1._create_json_hash(data=inner_list_siblings_hash_list, hash_function="sha256"), "value": 3},
            ],
            hash_function="sha256",
        )
        expected_out_hash_lists = [
            {"path": "[2]", "siblings": _jfpv1._create_json_hash(data=siblings_1, hash_function="sha256"), "value": 1},
            {"path": "[2]|[2]", "siblings": _jfpv1._create_json_hash(data=inner_list_siblings_hash_list, hash_function="sha256"), "value": 2},
            {"path": "[2]|[2]", "siblings": _jfpv1._create_json_hash(data=inner_list_siblings_hash_list, hash_function="sha256"), "value": 3},
        ]
        self.assertEqual(obj_out, expected_out_hash_lists)

        sorted_hash_list = _jfpv1._create_sorted_hash_list(data=obj_out, hash_function="sha256")
        hex_digest = _jfpv1._create_json_hash(data=sorted_hash_list, hash_function="sha256")
        jfpv1_out_1 = f"jfpv1$sha256${hex_digest}"
        jfpv1_out_2 = _jfpv1._create_jfpv1_fingerprint(data=obj_in, hash_function="sha256", version=1)
        self.assertEqual(jfpv1_out_1, jfpv1_out_2)

    def test_jfpv1_json_flattener_primitive_integer_handling(self):
        """Test jfpv1 primitive integer handling

        Verify that:
        - Integers are 'flattened' correctly (path, siblings and value)
        - Fingerprint matches with pre-verified fingerprint
        """
        int_val = 123
        int_out_raw = _jfpv1._flatten_json(data=int_val, hash_function="sha256")
        expected_int_out_raw = [{"path": "", "value": int_val}]
        self.assertEqual(int_out_raw, expected_int_out_raw)

    def test_jfpv1_json_flattener_primitive_float_handling(self):
        """Test jfpv1 primitive float handling

        Verify that:
        - Floats are 'flattened' correctly (path, siblings and value)
        - Fingerprint matches with pre-verified fingerprint
        """
        float_val = 123.321
        float_out_raw = _jfpv1._flatten_json(data=float_val, hash_function="sha256")
        expected_float_out_raw = [{"path": "", "value": float_val}]
        self.assertEqual(float_out_raw, expected_float_out_raw)

    def test_jfpv1_json_flattener_primitive_string_handling(self):
        """Test jfpv1 primitive string handling

        Verify that:
        - Strings are 'flattened' correctly (path, siblings and value)
        - Fingerprint matches with pre-verified fingerprint
        """
        string_val = "alpha 123"
        string_out_raw = _jfpv1._flatten_json(data=string_val, hash_function="sha256")
        expected_string_out_raw = [{"path": "", "value": string_val}]
        self.assertEqual(string_out_raw, expected_string_out_raw)

    def test_jfpv1_json_flattener_primitive_boolean_handling(self):
        """Test jfpv1 primitive boolean handling

        Verify that:
        - Booleans are 'flattened' correctly (path, siblings and value)
        - Fingerprint matches with pre-verified fingerprint
        """
        bool_val = True
        bool_out_raw = _jfpv1._flatten_json(data=bool_val, hash_function="sha256")
        expected_bool_out_raw = [{"path": "", "value": bool_val}]
        self.assertEqual(bool_out_raw, expected_bool_out_raw)

    def test_jfpv1_json_flattener_primitive_none_handling(self):
        """Test jfpv1 primitive boolean handling

        Verify that:
        - Booleans are 'flattened' correctly (path, siblings and value)
        - Fingerprint matches with pre-verified fingerprint
        """
        none_val = None
        none_out_raw = _jfpv1._flatten_json(data=none_val, hash_function="sha256")
        expected_bool_out_raw = [{"path": "", "value": none_val}]
        self.assertEqual(none_out_raw, expected_bool_out_raw)

    def test_jfpv1_json_flattener_empty_list_handling(self):
        """Test jfpv1 json flattener's ability to handle empty lists as values

        Versions up to v0.12.2 did not acknowledge empty lists as values.
        Related issue: https://github.com/cobaltine/json-fingerprint/issues/33

        Verify that:
        - Empty lists are considered to be values"""

        empty_list_val = []
        empty_list_out_raw = _jfpv1._flatten_json(data=empty_list_val, hash_function="sha256")
        expected_emtpy_list_out_raw = [{"path": "", "value": empty_list_val}]
        self.assertEqual(empty_list_out_raw, expected_emtpy_list_out_raw)

    def test_jfpv1_json_flattener_empty_dict_handling(self):
        """Test jfpv1 json flattener's ability to handle empty dicts as values

        Versions up to v0.12.2 did not acknowledge empty dicts as values.
        Related issue: https://github.com/cobaltine/json-fingerprint/issues/33

        Verify that:
        - Empty dicts are considered to be values"""

        empty_dict_val = {}
        empty_dict_out_raw = _jfpv1._flatten_json(data=empty_dict_val, hash_function="sha256")
        expected_empty_dict_out_raw = [{"path": "", "value": empty_dict_val}]
        self.assertEqual(empty_dict_out_raw, expected_empty_dict_out_raw)


if __name__ == "__main__":
    unittest.main()



class FingerprintHashFunctionError(Exception):
    pass


class FingerprintInputDataTypeError(Exception):
    pass


class FingerprintVersionError(Exception):
    pass


class FingerprintStringFormatError(Exception):
    pass


class FingerprintJSONLoadError(Exception):
    pass



import json
import unittest

from json_fingerprint import _exceptions, _find_matches, create


class TestFindMatches(unittest.TestCase):
    def setUp(self):
        self.test_input = json.dumps({"foo": "bar"})
        self.jfpv1_sha256 = create(input=self.test_input, hash_function="sha256", version=1)
        self.jfpv1_sha384 = create(input=self.test_input, hash_function="sha384", version=1)
        self.jfpv1_sha512 = create(input=self.test_input, hash_function="sha512", version=1)

    def test_get_target_hashes(self):
        """Test target hash list creation.

        Verify that:
        - The returned element list contains only unique entries
        """
        # Fingerprint list with duplicate entries
        fingerprints = [
            self.jfpv1_sha256,
            self.jfpv1_sha256,
            self.jfpv1_sha384,
            self.jfpv1_sha512,
            self.jfpv1_sha512,
        ]

        target_hashes = _find_matches._get_target_hashes(fingerprints=fingerprints)
        self.assertEqual(len(target_hashes), 3)

        expected_elements = [
            {"version": 1, "hash_function": "sha256"},
            {"version": 1, "hash_function": "sha384"},
            {"version": 1, "hash_function": "sha512"},
        ]
        self.assertEqual(expected_elements, target_hashes)

    def test_create_input_fingerprints(self):
        """Test list matching's fingerprint creation function.

        Verify that:
        - Fingerprints are correctly parsed from the given target hash elements
        """
        target_hashes = [
            {"version": 1, "hash_function": "sha256"},
            {"version": 1, "hash_function": "sha384"},
            {"version": 1, "hash_function": "sha512"},
        ]

        input_fingerprints = _find_matches._create_input_fingerprints(input=self.test_input, target_hashes=target_hashes)
        self.assertEqual(len(input_fingerprints), 3)
        self.assertIn(self.jfpv1_sha256, input_fingerprints)
        self.assertIn(self.jfpv1_sha384, input_fingerprints)
        self.assertIn(self.jfpv1_sha512, input_fingerprints)

    def test_jfpv1_find_matches(self):
        """Test json fingerprint list matcher.

        Verify that:
        - Fingerprints of all jfpv1 SHA-2 variants are properly matched in a fingerprint list
        - Deduplication works for duplicate entries in fingerprint list
        - Exceptions are properly raised with invalid fingerprints and input types
        """
        input = json.dumps({"bar": "foo"})
        chaff_jfpv1_sha256 = create(input=input, hash_function="sha256", version=1)

        # Fingerprint list with duplicate entries
        fingerprints = [
            # Expected matches (5)
            self.jfpv1_sha256,
            self.jfpv1_sha256,
            self.jfpv1_sha384,
            self.jfpv1_sha512,
            self.jfpv1_sha512,
            # Chaff
            chaff_jfpv1_sha256,
        ]

        matches = _find_matches.find_matches(input=self.test_input, fingerprints=fingerprints)
        self.assertEqual(len(matches), 5)

        deduplicated_matches = _find_matches.find_matches(input=self.test_input, fingerprints=fingerprints, deduplicate=True)
        self.assertEqual(len(deduplicated_matches), 3)

        with self.assertRaises(_exceptions.FingerprintJSONLoadError):
            _find_matches.find_matches(input='{"invalid": json string}', fingerprints=[self.jfpv1_sha256])
        with self.assertRaises(_exceptions.FingerprintStringFormatError):
            _find_matches.find_matches(input=input, fingerprints=["invalid fingerprint string"])


if __name__ == "__main__":
    unittest.main()



from typing import Dict, List

from ._create import create
from ._decode import decode


def _get_target_hashes(fingerprints: List[str]) -> List[Dict]:
    target_hashes = []
    for fingerprint in fingerprints:
        version, hash_function, _ = decode(fingerprint=fingerprint)
        element = {"version": version, "hash_function": hash_function}
        if element not in target_hashes:
            target_hashes.append(element)

    return target_hashes


def _create_input_fingerprints(input: str, target_hashes: List[Dict]) -> List[str]:
    input_fingerprints = []
    for element in target_hashes:
        fingerprint = create(input=input, hash_function=element["hash_function"], version=element["version"])
        input_fingerprints.append(fingerprint)
    return input_fingerprints


def find_matches(input: str, fingerprints: List[str], deduplicate: bool = False) -> List[str]:
    """Match raw json string input to a list of fingerprints

    Decodes the target fingerprints and creates a fingerprint from the input with identical parameters.
    Creates a fingerprint from the input of each different JSON fingerprint type present in the fingerprint list.
    """
    if deduplicate:
        fingerprints = list(set(fingerprints))
    target_hashes = _get_target_hashes(fingerprints=fingerprints)
    input_fingerprints = _create_input_fingerprints(input=input, target_hashes=target_hashes)

    matches = []
    for fingerprint in fingerprints:
        if fingerprint in input_fingerprints:
            matches.append(fingerprint)

    return matches



from typing import Tuple

from ._validators import _validate_fingerprint_format


def decode(fingerprint: str) -> Tuple[int, str, str]:
    """Decode json fingerprints into version, hash function and hash values."""
    _validate_fingerprint_format(fingerprint)
    elements = fingerprint.split("$")
    version = int(elements[0][4:])
    hash_function = elements[1]
    hash = elements[2]

    return version, hash_function, hash



import setuptools

with open("README.md", "r", encoding="utf-8") as f:
    long_description = f.read()

with open(".version", "r", encoding="utf-8") as f:
    version_raw = f.read()[1:]  # omit 'v'
    version = version_raw.split("-")[0]  # omit trailing version identifiers

setuptools.setup(
    name="json-fingerprint",
    version=version,
    author="Ville Lehtinen",
    author_email="ville.lehtinen@cobaltine.fi",
    license="MIT",
    description="Create consistent and comparable fingerprints with secure hashes from unordered JSON data",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/cobaltine/json-fingerprint",
    packages=setuptools.find_packages(),
    include_package_data=True,
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
    ],
    python_requires=">=3.8",
)



import json
import unittest

from json_fingerprint import _exceptions, create, match


class TestMatch(unittest.TestCase):
    def test_jfpv1_match(self):
        """Test json fingerprint matcher

        Verify that:
        - Fingerprints of all jfpv1 SHA-2 variants are properly matched
        - Exceptions are properly raised with invalid fingerprints and input types
        """
        input = json.dumps({"foo": "bar"})
        jfpv1_sha256 = create(input=input, hash_function="sha256", version=1)
        jfpv1_sha384 = create(input=input, hash_function="sha384", version=1)
        jfpv1_sha512 = create(input=input, hash_function="sha512", version=1)

        match_sha256 = match(input=input, target_fingerprint=jfpv1_sha256)
        self.assertEqual(match_sha256, True)
        match_sha384 = match(input=input, target_fingerprint=jfpv1_sha384)
        self.assertEqual(match_sha384, True)
        match_sha512 = match(input=input, target_fingerprint=jfpv1_sha512)
        self.assertEqual(match_sha512, True)

        no_match = match(input=json.dumps('{"bar": "foo"}'), target_fingerprint=jfpv1_sha256)
        self.assertEqual(no_match, False)

        with self.assertRaises(_exceptions.FingerprintJSONLoadError):
            match(input='{"invalid": json string}', target_fingerprint=jfpv1_sha256)
        with self.assertRaises(_exceptions.FingerprintStringFormatError):
            match(input=input, target_fingerprint="invalid fingerprint string")


if __name__ == "__main__":
    unittest.main()



from ._create import create
from ._decode import decode


def match(input: str, target_fingerprint: str) -> bool:
    """Match raw json string input to target fingerprint

    Decodes the target fingerprint and creates a fingerprint from the input with identical parameters."""
    version, hash_function, _ = decode(fingerprint=target_fingerprint)
    input_fingerprint = create(input=input, hash_function=hash_function, version=version)
    if input_fingerprint == target_fingerprint:
        return True
    return False
